package com.unitesk.reqcoverage.eclipse;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.viewers.CheckboxTreeViewer;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.wizard.WizardPage;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.model.WorkbenchLabelProvider;

import com.unitesk.reqcoverage.reqstracer.Collector;

public class PathToTestProject extends WizardPage {
	private CheckboxTreeViewer projectList;
	private final class ProjectListTreeProvider implements ITreeContentProvider {
		@Override
		public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		}

		@Override
		public void dispose() {
		}

		@Override
		public boolean hasChildren(Object element) {
			if(element == ResourcesPlugin.getWorkspace().getRoot()) {
				return true;
			}
			return false;
		}

		@Override
		public Object getParent(Object element) {
			if(element == ResourcesPlugin.getWorkspace().getRoot()) {
				return null;
			}
			return ResourcesPlugin.getWorkspace().getRoot();
		}

		@Override
		public Object[] getElements(Object inputElement) {
			return ResourcesPlugin.getWorkspace().getRoot().getProjects();
		}

		@Override
		public Object[] getChildren(Object parentElement) {
			if(parentElement == ResourcesPlugin.getWorkspace().getRoot()) {
				return getElements(parentElement);
			}
			return new Object[0];
		}
	}

//	private Text text;
	private Composite container;
	private List<IProject> storedProjects;
	
	public PathToTestProject() {
		super("Test projects");
		setTitle("Test projects");
		setDescription("Please select projects with tests");
		
	}

	@Override
	public void createControl(Composite parent) {
		container = new Composite(parent, SWT.NULL);
		container.setLayout(new GridLayout(1, false));
		projectList = new CheckboxTreeViewer(container);
		projectList.getControl().setLayoutData(new GridData(GridData.FILL_BOTH));
		projectList.setContentProvider(new ProjectListTreeProvider());
		projectList.setLabelProvider(new WorkbenchLabelProvider());
		projectList.setInput(ResourcesPlugin.getWorkspace().getRoot());
		if(storedProjects!=null)
			for(IProject p: storedProjects)
				projectList.setChecked(p, true);
		
		
		setControl(container);
		Collector.instance().setProcessReport(false);
		
	}
	
	public ArrayList<IProject> getSelectedProjects() {
		ArrayList<IProject> result = new ArrayList<IProject>();
		for(Object p: projectList.getCheckedElements()){
			result.add((IProject) p);
		}
		return result;
	}

	public void setSelectedProjects(List<IProject> projs) {
		storedProjects = projs;
		if(projectList != null){
			for(IProject p : projs)
				projectList.setChecked(p, true);
		}
	}
}

