/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.reqcoverage.handlers;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.Command;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.IHandler;
import org.eclipse.core.commands.NotEnabledException;
import org.eclipse.core.commands.NotHandledException;
import org.eclipse.core.commands.common.NotDefinedException;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.ICommandService;
import org.eclipse.ui.dialogs.ElementListSelectionDialog;
import org.eclipse.ui.dialogs.ListSelectionDialog;
import org.eclipse.ui.services.IServiceLocator;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.ui.cnf.RequalityCNF;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.tests.CoverageStorage;
import com.unitesk.requality.tests.ICoverageSourceWizard;
import com.unitesk.requality.tests.ICoverageStorageSource;
// TODO report undo
public class SetCoverageSource extends AbstractHandler implements IHandler {
	@Override
	public boolean isEnabled() {
		List<TreeNode> nodes = RequalityCNF.getSelectedNodes();
        if(nodes == null)
            return false;
        if( nodes.size() != 1)
        	return false;
        if(!nodes.get(0).getType().equals(ReportData.TYPE_NAME))
        	return false;
        String s = ((ReportData)nodes.get(0)).getTemplateId();
        if(s==null || !s.equals("com.unitesk.requality.reports.coverage"))
        	return false;
        return true;
	};
	
	private class IdToName{
		private String id, name;
		public IdToName(String id, String name){
			this.id=id;
			this.name=name;
		}
		public String getName(){
			return name;
		}
		public String getId(){
			return id;
		}
		@Override
		public String toString() {
			return name;
		}
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof IdToName)
				return ((IdToName)obj).getId().equals(id);
			return super.equals(obj);
		}
	}
	
    @Override
    public Object execute(ExecutionEvent event) throws ExecutionException {
        List<TreeNode> nodes = RequalityCNF.getSelectedNodes();
        TreeNode node = nodes.get(0);
        try{
        	
        	ElementListSelectionDialog dialog = new ElementListSelectionDialog(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), new LabelProvider());

        	dialog.setTitle("Coverage Source Selection");

        	dialog.setMessage("Select a Source of Coverage Information");

        	List<Object> elems = new ArrayList<Object>();
         	IdToName idtm = new IdToName("com.unitesk.requality.defaultsource","From File");
        	elems.add(idtm);
        	Map<String,String> map = Activator.getDefault().getCoverageSourceNames();
        	
        	String selnow = node.getStringAttributeValue("coverageStorageType");
        	IdToName temp = null;
        	for(String s:map.keySet()){
        		IdToName o = new IdToName(s, map.get(s));
        		elems.add(o);
        		if(selnow!=null && s.equals(selnow))
        			temp = o;
        	}
        	dialog.setMultipleSelection(false);
        	dialog.setElements(elems.toArray());
        	List sel = new ArrayList();
        	if(temp != null)
        		sel.add(temp);
        	else
        		sel.add(idtm);
        	dialog.setInitialElementSelections(sel);
        	int res = dialog.open();
        	if(res != ListSelectionDialog.OK)
        		return null;
        	Object[] result = dialog.getResult();
        	if(result[0].equals(idtm)){
        		FileDialog fd = new FileDialog(getShell());
        		fd.setFilterExtensions(new String[]{"*.xml"});
        		fd.setFilterNames(new String[]{"Coverage Information File"});
        		String oldpath = node.getStringAttributeValue("coverageFilePath");
        		if(oldpath!=null)
        			fd.setFileName(oldpath);
        		String fname = fd.open();
        		if(fname!=null){
        			try{
        				node.getTreeDB().startTransaction("Change coverage source");
        				Attribute a = node.getAttribute("coverageStorageSettings");
        				if(a!=null)
        					node.removeAttribute(a.getKey());
        				a = node.getAttribute("coverageStorageType");
        				if(a!=null)
        					node.removeAttribute(a.getKey());
        				node.putAttribute(new Attribute(AttributeType.STRING,"coverageFilePath",fname));
        				node.saveAttributes();
        				node.getTreeDB().commit();
        			}
        			catch(Exception e){
        				node.getTreeDB().rollback();
        			}
        		}
        		else
         			return null;
    			
        	}else{
				Shell shell = getShell();
				String id = ((IdToName)result[0]).getId();
				ICoverageStorageSource source = Activator.getCoverageSource(id);
				ICoverageSourceWizard wiz = source.getWizard(node.getTreeDB());
				Attribute tp = node.getAttribute("coverageStorageType");
				Attribute set = null;
				if(tp != null && tp.getValue().equals(id))
					set = node.getAttribute("coverageStorageSettings");
    			if (wiz == null)
    				return null;
    			if(set != null)
    				wiz.setCoverageStorageSettings(((String)(set.getValue())).split(","));
    			WizardDialog wd = new WizardDialog(shell, wiz);
    			if (wd.open() == WizardDialog.OK){
    				try{
    					node.getTreeDB().startTransaction("Change coverage source");
    					Attribute a = node.getAttribute("coverageFilePath");
    					if(a!=null)
    						node.removeAttribute(a.getKey());
    					CoverageStorage cs = CoverageStorage.getStorage(node.getTreeDB(), source);
    					node.putAttribute(new Attribute(AttributeType.STRING,"coverageStorageType",id));
    					node.putAttribute(new Attribute(AttributeType.STRING,"coverageStorageSettings",CoreUtils.listToString(cs.getSettings(), ",")));
    					node.saveAttributes();
    					node.getTreeDB().commit();
    				}
        			catch(Exception e){
        				node.getTreeDB().rollback();
        			}
    			}else
    				return null;
        	}
        	
        	
        	IServiceLocator serviceLocator = PlatformUI.getWorkbench();
        	ICommandService commandService = (ICommandService) serviceLocator.getService(ICommandService.class);
        	try  { 
        	    Command command = commandService.getCommand("com.unitesk.requality.commands.GenerateReport");
        	    ExecutionEvent ev = new ExecutionEvent();
        	    command.executeWithChecks(ev);
        	} catch (ExecutionException e) {
        	    e.printStackTrace();
        	}
        	catch(NotDefinedException e2){
        		e2.printStackTrace();
        	}
        	catch(NotEnabledException e3){
        		e3.printStackTrace();
        	}
        	catch(NotHandledException e4){
        		e4.printStackTrace();
        	}
        	
        }
        catch(Exception e){
        	e.printStackTrace();
        	TreeDB db = node.getTreeDB();
        	if(db!=null && db.getStorage() instanceof TransactionStorage)
        		db.rollback();
        }
        return null;
    }
    private static Shell getShell(){
    	if(Activator.getDefault() == null)
    		return null;
    	IWorkbench w = Activator.getDefault().getWorkbench();
    	if(w==null)
    		return null;
    	IWorkbenchWindow win = w.getActiveWorkbenchWindow();
    	if(win == null)
    		return null;
    	return win.getShell();
    }
}
