package com.unitesk.reqcoverage.tests;

import java.io.File;
import java.io.IOException;
import java.security.acl.LastOwnerException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;

import com.unitesk.reqcoverage.Activator;
import com.unitesk.reqcoverage.eclipse.GetPathToTestProjects;
import com.unitesk.reqcoverage.reqstracer.Collector;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.tests.CoverageStorage;
import com.unitesk.requality.tests.ICoverageSourceWizard;
import com.unitesk.requality.tests.ICoverageStorageSource;
import com.unitesk.requality.tests.TestErrorInfo;
import com.unitesk.requality.tests.TreeNodeCoverageInfo;

public class CoverageByProjects implements ICoverageStorageSource {
	public static final String ID = "com.unitesk.reqcoverage.fromProject";
	public List<String> extension;
	public List<String> fileregexp;  
	public Collector collector;
	private String lastErrors;
	private List<IProject> projects;
	private CoverageStorage targetCoverageStorage;
	
	public void setCollector(Collector instance){
		collector = instance;
		projects = instance.getTestProjectsList();
		fileregexp = instance.getRegExpressions();
		extension = instance.getExtensions();
	}
	public CoverageByProjects(){
		
	}
	
	public CoverageByProjects(Collector instance) {
		setCollector(instance);
	}
	private Collector getCollector(){
		if(collector == null){
			collector = Collector.instance();
			collector.setProcessReport(true);
			collector.setTestProjectsList(new ArrayList<IProject>(projects));
			collector.setExtensions(new ArrayList<String>(extension));
			collector.setRegExpressions(new ArrayList<String>(fileregexp));			
		}
		return collector;
	}
	@Override
	public Map<String, TreeNodeCoverageInfo> getCoverage() {
		try {
			if(collector != null)
				collector.reset(); 
			collector = null;

			getCollector().selectFilesToFindReqs();
			getCollector().constructMapReqsToFiles();
		} catch (CoreException e) {
			lastErrors = e.getMessage();
			e.printStackTrace();
		} catch (IOException e) {
			lastErrors = e.getMessage();
			e.printStackTrace();
		}
		Map<String, TreeNodeCoverageInfo> res = new HashMap<String, TreeNodeCoverageInfo>();
		Map<String, ArrayList<String>> map = getCollector().getReqsToFilesMap();
		TreeDB db = targetCoverageStorage.getTreeDB();
		for(String s:map.keySet()){
			String bases = s;
			if(db!=null)
			{
				try{
					s = db.getNode(s).getQualifiedId();
				}
				catch(Exception e){
				}
			}
			TreeNodeCoverageInfo tnci = new TreeNodeCoverageInfo();
			tnci.setTarget(s);
			for(String s2:map.get(bases))
				tnci.registerTest("file:"+IPath.SEPARATOR+IPath.SEPARATOR+IPath.SEPARATOR+s2, 1);
			res.put(s, tnci);
		}
		return res;
	}
	@Override
	public void setSettings(String... setts) {
		if(collector != null)
			collector.reset();
		collector = null;
		fileregexp = new ArrayList<String>();
		projects = new ArrayList<IProject>();
		extension = new ArrayList<String>();
		if(setts.length>=3){
			fileregexp =  Arrays.asList(setts[0].split("\\|"));
			projects = Activator.getProjectsByNames(setts[1].split("\\|"));
			extension =  Arrays.asList(setts[2].split("\\|"));
		}
		getCollector();
	}
	@Override
	public String getLastErrors() {
		return lastErrors;
	}
	@Override
	public String[] getSettings() {
		String[] arr = new String[3];
		arr[0] = CoreUtils.listToString(fileregexp, "|");
		arr[1] = "";
		for(IProject p:projects)
			arr[1]+=p.getName()+"|";
		if(!arr[1].equals(""))
			arr[1] = arr[1].substring(0,arr[1].length()-1);
		arr[2] = CoreUtils.listToString(extension, "|");
		return arr;
	}
	@Override
	public Map<String, TestErrorInfo[]> getErrors() {
		return new HashMap<String, TestErrorInfo[]>();
	}
	@Override
	public void setTargetStorage(CoverageStorage coverageStorage) {
		targetCoverageStorage = coverageStorage;
	}
	@Override
	public ICoverageSourceWizard getWizard(TreeDB db) {
		ICoverageSourceWizard wiz = new GetPathToTestProjects(this);
		wiz.setCoverageStorageSettings(new String[]{"\\* @requirement (.*)","","ttcn"});
		return wiz;
	}

}
